/************************************************************************
*
* lnklibrary - Loads library objects from asm980 for linking.
*
* Changes:
*   06/29/11   DGP   Original. Hacked from lnk990.
*
************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>

#include "lnkdef.h"

extern int pc;

extern uint16 memory[MEMSIZE];
extern Memory memctl[MEMSIZE];


/************************************************************************
* checkdef - Check if DEF is needed.
************************************************************************/

static int
checkdef (char *item, int relo, int flags)
{
   SymNode *s;
   char *bp;

   bp = item;
   for (; *bp; bp++) if (isspace(*bp)) *bp = '\0';
#ifdef DEBUGLIBRARY
   printf (", %c%s = %s",
	 relo ? 'R' : 'A', "DEF", item);
#endif
   if ((s = symlookup (item, "", FALSE)) != NULL)
   {
      if (s->flags & UNDEF && !(s->flags & SREF))
      {
#ifdef DEBUGLIBRARY
	 printf ("\n      flags = %04X, value = %04X",
	       s->flags, s->value);
#endif
         return (1);
      }
   }
   return (0);
}

/************************************************************************
* lnkcheck - Check if file is referenced.
************************************************************************/

static int
lnkcheck (FILE *fd)
{
   int status;
   int reclen;
   int relo;
   int done;
   int i;
   int wordlen = WORDTAGLEN-1;
   unsigned char inbuf[82];

#ifdef DEBUGLIBRARY
   printf ("lnkcheck: \n");
#endif

   status = 0;
   i = fgetc (fd);
   ungetc (i, fd);
   if (i != IDT_TAG)
   {
      fprintf (stderr, "lnk980: Card format not supported in libraries\n");
      exit (ABORT);
   }


   done = FALSE;
   while (!done)
   {
      unsigned char *op = inbuf;

      if (fgets ((char *)inbuf, 82, fd) == NULL)
      {
	 done = TRUE;
	 break;
      }
      reclen = strlen ((char *)inbuf);

#ifdef DEBUGLIBRARY
      printf ("reclen = %d\n", reclen);
#endif
      if (*op == EOFSYM)
      {
	 done = TRUE;
	 break;
      }

      for (i = 0; i < reclen; i++)
      {
	 char otag;
	 char item[80];

	 otag = *op++;

#ifdef DEBUGLIBRARY
	 {
	    uint16 wdata;
	    wdata = getnumfield (op, FALSE);
	    printf ("   otag = %c", isprint(otag) ? otag : '0');
	    printf (", data = %04X", wdata & 0xFFFF);
	 }
#endif

	 relo = FALSE;
	 switch (otag)
	 {
	 case IDT_TAG:
	    op += wordlen;
	    strncpy (item, (char *)op, IDTSIZE);
	    item[IDTSIZE] = '\0';
#ifdef DEBUGLIBRARY
	    printf (", IDT = %s", item);
#endif
	    op += IDTSIZE;
	    break;

	 case RELEXTRN_TAG:
	 case ABSEXTRN_TAG:
	    op += wordlen;
	    op += MAXSYMLEN;
	    break;

	 case RELGLOBAL_TAG:
	    relo = TRUE;
	 case ABSGLOBAL_TAG:
	    op += wordlen;
	    strncpy (item, (char *)op, MAXSYMLEN);
	    item[MAXSYMLEN] = '\0';
	    status += checkdef (item, relo, 0);
	    op += MAXSYMLEN;
	    break;

	 case EOR_TAG:
	    i = 81;
	    break;

	 case NOCKSUM_TAG:
	 case CKSUM_TAG:
	    op += wordlen;
	    break; 

	 default:
	    op += wordlen;
	 }
#ifdef DEBUGLIBRARY
	 printf ("\n");
#endif
      }
   }

   return (status);

}

/************************************************************************
* lnklibrary - Object library loader for lnk980.
************************************************************************/

int
lnklibrary (FILE *fd, int loadpt, int pass, char *file)
{
   long modstart;
   int more;
   int pcstart;
   int curraddr;

   curraddr = loadpt;

   /*
   ** Process the entire library
   */

   do
   {
      fseek (fd, 0, SEEK_SET);
      clearerr(fd);
      more = FALSE;

      while (!feof(fd))
      {
	 modstart = ftell(fd);
	 pcstart = curraddr;

	 /*
	 ** Check if current module has DEFs that are needed for an
	 ** unresolved reference.
	 */

	 if (lnkcheck (fd))
	 {
	    int i;

	    /*
	    ** Yes, load module.
	    */

	    more = TRUE;
	    if (pcstart & 1) pcstart ++;
	    fseek (fd, modstart, SEEK_SET);
	    curraddr = lnkloader (fd, pcstart, 1, file, TRUE);
	    memset (&memory[pcstart], 0, curraddr-pcstart);
	    for (i = pcstart; i < curraddr; i++)
	    {
	       memset (&memctl[i], '\0', sizeof (Memory));
	       memctl[i].tag = ABSDATA_TAG;
	    }
	    fseek (fd, modstart, SEEK_SET);
	    curraddr = lnkloader (fd, pcstart, 2, file, TRUE);
	 }
      }
   } while (more);
   pc = curraddr;
   return (0);
}
