/************************************************************************
*
* lnkpunch - Punchs out memory.
*
* Changes:
*   06/29/11   DGP   Original. Hacked from lnk990.
*   08/04/11   DGP   Added 980 card object format.
*
************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>

#include "lnkdef.h"

extern FILE *lstfd;
extern int listmode;
extern int pc;
extern int errcount;
extern int absentry;
extern int cardoutmode;
extern int relentry;
extern int symbolcount;

extern uint16 memory[MEMSIZE];
extern Memory memctl[MEMSIZE];
extern char idtbuf[IDTSIZE+2];
extern SymNode *symbols[MAXSYMBOLS];
extern struct tm *timeblk;

static int objcnt = 0;
static int objrecnum = 0;
static int cardword = 0;
static int cardblock = 0;
static int pgmlength;

static uint16 reloword = 0;

static char objbuf[80];
static char objrec[82];
static char cardrec[80];

/***********************************************************************
* punchfinish - Punch a record with sequence numbers.
***********************************************************************/

static void
punchfinish (FILE *outfd)
{
   if (objcnt)
   {
      if (!cardoutmode)
      {
	 int i;
	 short int cksum;
	 char temp[10];

	 cksum = 0;
	 for (i = 0; i < objcnt; i++) cksum += objrec[i];
	 cksum += CKSUM_TAG;
	 cksum = -cksum;
	 sprintf (temp, "%c%04X%c", CKSUM_TAG, cksum & 0xFFFF, EOR_TAG);
	 strncpy (&objrec[objcnt], temp, 6);
	 objcnt += 6;

	 if (idtbuf[0])
	    strncpy (&objrec[IDTCOLUMN], idtbuf, strlen(idtbuf));
      }
      sprintf (&objrec[SEQUENCENUM], SEQFORMAT, ++objrecnum);
      fputs (objrec, outfd);
      memset (objrec, ' ', sizeof(objrec));
      objcnt = 0;
   }
}

/***********************************************************************
* punchrecord - Punch an object value into record.
***********************************************************************/

static void 
punchrecord (FILE *outfd)
{
   int len = strlen (objbuf);

   if (cardoutmode)
   {
      char *op = objbuf;
      int relo = 0;
      char temp[16];

      len--;
      switch (*op++)
      {
      case RELDATA_TAG:
         relo = 2;
      case ABSDATA_TAG:
	 if (cardblock == 0)
	 {
	    cardblock++;
	    sprintf (objrec, "%04X", pc);
	    objcnt = 4;
	 }
	 reloword |= relo << (14 - (cardword*2));
	 sprintf (temp, "%04X", reloword);
	 strncpy (cardrec, temp, 4);
	 strncpy (&cardrec[4+(cardword*4)], op, len);
	 strncpy (&objrec[objcnt], cardrec, 8+(cardword*4));
	 cardword++;
	 if (cardword > 7)
	 {
	    memset (cardrec, ' ', sizeof(cardrec));
	    if (cardblock == 2)
	       cardblock = 0;
	    else
	       cardblock++;
	    objcnt += 36;
	    cardword = 0;
	    reloword = 0;
	    if (objcnt >= 74)
	       punchfinish (outfd);
	 }
	 break;

      case RELENTRY_TAG:
      case ABSENTRY_TAG:
	 punchfinish (outfd);
	 sprintf (cardrec, "&&&&%s", op);
	 strncpy (objrec, cardrec, 8);
	 objcnt += 8;
	 punchfinish (outfd);
	 break;

      case RELORG_TAG:
      case ABSORG_TAG:
         punchfinish (outfd);
	 memset (cardrec, ' ', sizeof(cardrec));
	 cardblock = 0;
	 cardword = 0;
	 reloword = 0;
	 break;

      default: ;
      }
   }
   else
   {
      if (objcnt+len >= CHARSPERREC)
      {
	 punchfinish (outfd);
      }
      strncpy (&objrec[objcnt], objbuf, len);
      objcnt += len;
   }
   objbuf[0] = '\0';
}

/***********************************************************************
* punchsymbols - Punch REF and DEF symbols.
***********************************************************************/

static void
punchsymbols (FILE *outfd)
{
   SymNode *sym;
   int i;

   if (cardoutmode)
      return;

   for (i = 0; i < symbolcount; i++)
   {
      sym = symbols[i];

      if (sym->flags & EXTERNAL)
      {
	 sprintf (objbuf, REFFORMAT,
		  (sym->flags & RELOCATABLE) ? RELEXTRN_TAG : ABSEXTRN_TAG,
		  sym->value & 0xFFFF, sym->symbol);
         
         punchrecord (outfd);
      }
      else if (sym->flags & GLOBAL)
      {
	 sprintf (objbuf, DEFFORMAT,
		  (sym->flags & RELOCATABLE) ? RELGLOBAL_TAG : ABSGLOBAL_TAG,
		  sym->value & 0xFFFF, sym->symbol);
         punchrecord (outfd);
      }
   }
}

/***********************************************************************
* puncheof - Punch EOF mark.
***********************************************************************/

static void
puncheof (FILE *outfd)
{
   char temp[80];

   punchfinish (outfd);
   if (cardoutmode)
   {
      objrec[0] = '/';
      objrec[1] = '*';
   }
   else
   {
      objrec[0] = EOFSYM;
      if (idtbuf[0])
	 strncpy (&objrec[IDTCOLUMN], idtbuf, strlen(idtbuf));
   }
   sprintf (temp, "%-8.8s  %02d/%02d/%02d  %02d:%02d:%02d    LNK980 %s",
	    idtbuf,
	    timeblk->tm_mon+1, timeblk->tm_mday, timeblk->tm_year - 100,
	    timeblk->tm_hour, timeblk->tm_min, timeblk->tm_sec,
	    VERSION);
   strncpy (&objrec[7], temp, strlen(temp));
   sprintf (&objrec[SEQUENCENUM], SEQFORMAT, ++objrecnum);
   fputs (objrec, outfd);
}

/***********************************************************************
* punchcardstart - Punch a card format preamble.
***********************************************************************/

static void
punchcardstart (FILE *outfd)
{
#ifdef DEBUGPUNCH
   printf ("punchcardstart: IDT = %s, length = %04X\n", idtbuf, pgmlength);
#endif

   objcnt = sprintf (objbuf, CARDIDTFORMAT,
		     idtbuf, 1, pgmlength, 0, 0, 0);
   strncpy (objrec, objbuf, objcnt);
   punchfinish (outfd);
}

/***********************************************************************
* lnkpunch - Punch out memory.
***********************************************************************/

int
lnkpunch (FILE *outfd, int partiallink)
{
   pgmlength = pc;

#ifdef DEBUGPUNCH
   printf ("lnkpunch: pgmlength = %04X\n", pgmlength);
#endif
   memset (objrec, ' ', sizeof(objrec));

   if (cardoutmode)
   {
      punchcardstart (outfd);
   }
   else
   {
      sprintf (objbuf, IDTFORMAT, IDT_TAG, pgmlength, idtbuf);
      punchrecord (outfd);
      sprintf (objbuf, OBJFORMAT, RELORG_TAG, 0);
      punchrecord (outfd);
   }

   for (pc = 0; pc < pgmlength;)
   {
#ifdef DEBUGPUNCH
      printf ("   pc = %04X, tag = %c(%2X), data = %04X\n",
      	      pc, memctl[pc].tag, memctl[pc].tag, GETMEM(pc));
#endif
      if (memctl[pc].tag)
      {
         sprintf (objbuf, OBJFORMAT, memctl[pc].tag, GETMEM(pc));
         punchrecord (outfd);
	 if (memctl[pc].tag == RELORG_TAG)
	 {
	    int j = pc;
	    pc = GETMEM(pc);
	    if (pc == j) pc++;
	 }
	 else pc++;
      }
   }

   if (relentry >= 0)
   {
      sprintf (objbuf, OBJFORMAT, RELENTRY_TAG, relentry);
      punchrecord (outfd);
   }

   if (partiallink) punchsymbols (outfd);

   puncheof (outfd);

   return (0);
}

